module HhowmmReminder
    (
      ReminderItem(..)
    , ReminderInfo(..)
    , searchReminders
    , priority_bottom
    , priority_invisible
    ) where

import Control.Applicative hiding (many, optional)
import Control.Monad
import Data.Char
import Data.Either
import Data.Maybe
import Data.Time
import Data.Time.Format
import System.Directory
import System.Locale

import Text.Parsec
import Text.Parsec.String

import HhowmmFilename
import HhowmmGrep

data ReminderInfo = ReminderInfo
    { dueDate :: Day
    , dueTime :: Maybe TimeOfDay
    , reminderType :: Char
    , priority :: Int
    } deriving (Show, Eq)
data ReminderItem = ReminderItem
    { grepItem :: GrepItem
    , info :: ReminderInfo
    } deriving (Show, Eq)

instance Ord ReminderInfo where
    compare a b = compare (priority a) (priority b)
instance Ord ReminderItem where
    compare a b = compare (info a) (info b)

priority_bottom :: Int
priority_bottom = 65534
priority_invisible :: Int
priority_invisible = 65535

nDigit :: Parser Int
nDigit = digitToInt <$> digit

nsToInt :: [Int] -> Int
nsToInt = foldl (\x y -> x * 10 + y) 0

digitByte :: Parser Int
digitByte = nsToInt <$> many1 nDigit

maybeTime :: Parser (Maybe TimeOfDay)
maybeTime = do
    str <- optionMaybe (many (noneOf "]"))
    return ((str >>= parseTime defaultTimeLocale "%H:%M") :: Maybe TimeOfDay)
--  case str of
--      Nothing -> return Nothing
--      Just t  -> return (parseTime defaultTimeLocale "%H:%M" t :: Maybe TimeOfDay)

reminderLine :: String -> LocalTime -> Parser ReminderInfo
reminderLine format now = do
    char '['
    str <- many (noneOf "\t ]")
    let date = parseTime defaultTimeLocale format str :: Maybe Day
    when (isNothing date) (unexpected "Illegal date format")
    optional (many1 space)
    time <- maybeTime
    char ']'
    optional (many1 space)
    rType <- oneOf "@!-~"
    arg <- option (-1) digitByte
    return $ ReminderInfo (fromJust date) time rType (calcPriority (fromJust date) time rType arg now)

calcPriority :: Day -> Maybe TimeOfDay -> Char -> Int -> LocalTime -> Int
calcPriority date _ '@' arg now
    | pr + arg < 0  = priority_invisible
    | pr < 0        = 0
    | otherwise     = pr
  where pr = fromIntegral $ date `diffDays` localDay now
calcPriority date time '!' arg now
    | today > date                                          = fromIntegral $ date `diffDays` today
    | today > date' || (today == date' && nowTime > time')  = 0
    | otherwise                                             = fromIntegral $ date' `diffDays` today
  where
    today   = localDay now
    nowTime = localTimeOfDay now
    date'   = addDays (fromIntegral (-arg)) date
    time'   = fromMaybe (TimeOfDay 23 59 59) time
calcPriority date time '-' arg now
    | today > date'                                         = fromIntegral $ today `diffDays` date'
    | today > date || (today == date && nowTime > time')    = 0
    | otherwise                                             = fromIntegral $ date `diffDays` today
  where
    today   = localDay now
    nowTime = localTimeOfDay now
    date'   = addDays (fromIntegral arg) date
    time'   = fromMaybe (TimeOfDay 0 0 0) time
calcPriority date _ '~' arg now
    | date > today                      = fromIntegral $ date `diffDays` today
    | otherwise                         = limitHalf arg $ (fromIntegral $ today `diffDays` date) `mod` arg
  where today     = localDay now
calcPriority _ _ _ _ _ = priority_invisible

limitHalf :: Int -> Int -> Int
limitHalf x y
    | (fromIntegral y :: Double) > (fromIntegral x :: Double)/2.0  = x-y
    | otherwise = y


parseReminderLine :: String -> LocalTime -> Either ParseError ReminderInfo
parseReminderLine lineStr now = parse (reminderLine "%Y-%m-%d" now) "(test)" lineStr

search :: String -> LocalTime -> [(Int, String, ReminderInfo)]
search content now =
    [(a, b, c) | (a, b, Right c) <- results]
  where
    linesWithNumber = zip [0..] $ lines content
    results = map (\(num, lineStr) -> (num, lineStr, parseReminderLine lineStr now)) linesWithNumber

getLocalTime :: IO LocalTime
getLocalTime = do
    now <- getZonedTime
    return $ zonedTimeToLocalTime now

processFile :: FilePath -> IO [ReminderItem]
processFile file = do
    exists <- doesFileExist file
    now <- getLocalTime
    if not exists
        then return []
        else do
            content <- readFile file
            return $ map (\(n, str, rInfo) -> ReminderItem (GrepItem file n str) rInfo) (search content now)

searchReminders :: [FilePath] -> IO [ReminderItem]
searchReminders files = do
    results <- forM files processFile
    return $ concat results
