module HhowmmResults
    (
      resultsWindow
    ) where

import Control.Monad (when, unless)
import Data.Bits
import Data.Char
import Data.List
import Data.Maybe
import System.Process

import Graphics.UI.WX hiding (when)
import Graphics.UI.WXCore hiding (when)

import HhowmmConfig
import HhowmmFilename
import HhowmmGrep
import HhowmmDefs

data GUIContext = GUIContext
    { guiFrame :: Frame ()
    , guiTitlesView :: ListView GrepItem
    , guiFileNameView :: StaticText ()
    , guiContentView :: TextCtrl ()
    , textAttrs :: [(String, TextAttr ())]
    , config :: ConfigInfo
    , patterns :: [String]
    }

attr_name_matched   :: String
attr_name_matched   = "matched"
attr_name_title     :: String
attr_name_title     = "title"
attr_name_comefrom  :: String
attr_name_comefrom  = "comefrom"

attrs_list :: [(String, Color, Color, Int)]
attrs_list = [(attr_name_matched, color_null, color_lightblue, wxFONTWEIGHT_NORMAL), (attr_name_title, color_darkblue, color_null, wxFONTWEIGHT_BOLD), (attr_name_comefrom, color_darkgreen, color_null, wxFONTWEIGHT_NORMAL)]

searchWords :: [String] -> String -> [(Int, Int)]
searchWords pats str =
    concatMap (\pat -> map (\x -> (x, x+(length pat))) $ findIndices (isPrefixOf pat) $ tails str) pats

searchTitles :: String -> [(Int, Int)]
searchTitles str =
    map (\(x, y) -> (y, y+(length x)))
        $ filter (\(x, _) -> isPrefixOf title_str x)
        $ zip listOfLines
        $ scanl (\x y -> x+(length y)+1) 0 listOfLines
  where
    listOfLines = lines str

searchLinks :: String -> [(Int, Int)]
searchLinks str =
    map (\x -> (x, x+(findIndexFrom x isSpace str))) $ findIndices (isPrefixOf "http://") $ tails str
  where
    findIndexFrom n p list = fromMaybe (length list) (findIndex p $ drop n list)


listHandler :: GUIContext -> EventList -> IO ()
listHandler _   (ListItemSelected (-1))     = return ()
listHandler ctx (ListItemSelected    n)     = do
    its <- listViewGetItems (guiTitlesView ctx)
    loadText (its !! n) ctx
listHandler ctx (ListKeyDown (KeyChar 'E')) = do
    let editor = confEditorCmd $ config ctx
    itemSelected <- listCtrlGetNextItem (listViewCtrl $ guiTitlesView ctx) (-1) wxLIST_NEXT_ALL wxLIST_STATE_SELECTED
    when (itemSelected /= (-1)) $ do
        its <- listViewGetItems (guiTitlesView ctx)
        (_, _, _, pHandle) <- createProcess (shell $ editor ++ " " ++ (fileName $ its !! itemSelected))
        waitForProcess pHandle
        loadText (its !! itemSelected) ctx
        return ()
listHandler _ _                             = propagateEvent

loadText :: GrepItem -> GUIContext -> IO ()
loadText it context = do
    let tv = guiContentView context
    textCtrlLoadFile tv (fileName it)

    let matchedAttr = find (\attr -> fst attr == attr_name_matched) (textAttrs context)
    let titleAttr   = find (\attr -> fst attr == attr_name_title) (textAttrs context)
    unless (isNothing matchedAttr || isNothing titleAttr) $ do
        contentText <- get tv text

        applyAttr matchedAttr (searchWords (patterns context) contentText) tv
        applyAttr titleAttr (searchTitles contentText) tv

        pos <- textCtrlXYToPosition tv $ Point 0 (lineNum it)
        textCtrlShowPosition tv pos

        set (guiFileNameView context) [text := fileName it]
  where
    applyAttr attr regions tv = mapM_ (\(st, en) -> textCtrlSetStyle tv st en (snd $ fromJust attr)) regions

resultsWindow :: ConfigInfo -> [String] -> [FilePath] -> IO ()
resultsWindow conf pats files = do
    f <- frame [text := "Hhowmm search results"]

    titlesView      <- listView f [""] (\it -> [lineText it])
    its <- if (length pats == 1)
            then if isPrefixOf "#" (head pats)
                then hashtagSearch (head pats) files >>= sortByModTime
                else andGrep pats files >>= sortByModTime
            else andGrep pats files >>= sortByModTime
    listViewSetItems titlesView its
    set (listViewCtrl titlesView) [ font := fontDefault{_fontSize = confFontSize conf}
                                  , style :~ (.|. (wxLC_NO_HEADER .|. wxLC_SINGLE_SEL))
                                  , on escKey := close f
                                  ]

    textView        <- textCtrlRich f [ wrap := WrapWord
                                      , font := fontDefault{_fontSize = confFontSize conf}
                                      , style :~ (.|. wxTE_AUTO_URL)
                                      , on escKey := windowSetFocus (listViewCtrl titlesView)
                                      ]
    textCtrlSetEditable textView False
    attrs <- mapM (\(name, fg, bg, weight) -> do
                        fnt <- fontCreateDefault
                        fontSetPointSize fnt (confFontSize conf)
                        fontSetWeight fnt weight
                        attr <- textAttrCreateDefault
                        textAttrSetFont attr fnt
                        unless (fg == color_null) $ textAttrSetTextColour attr fg
                        unless (bg == color_null) $ textAttrSetBackgroundColour attr bg
                        return (name, attr)
                    ) attrs_list

    fileNameView    <- staticText f []

    let context     = GUIContext f titlesView fileNameView textView attrs conf pats
    listViewSetHandler titlesView (listHandler context)

    set f [layout :=
            margin 10 $
            column 2
                [ hfill $ listViewLayout titlesView
                , hfill $ widget fileNameView
                , fill $ widget textView
                ]
          , clientSize := sz 400 800
          ]

    wSize <- windowGetSize f
    listCtrlSetColumnWidth (listViewCtrl titlesView) 0 (sizeW wSize)
    windowSetFocus (listViewCtrl titlesView)

    unless (null its) $ listHandler context (ListItemSelected 0)
