module Main where

import Control.Applicative
import Control.Monad (unless)
import Data.Bits
import Data.List
import Data.Maybe (fromMaybe, isNothing)
import System.Directory
import System.Environment
import System.FilePath
import System.PosixCompat.Files
import System.Process

import Graphics.UI.WX
import Graphics.UI.WXCore

import HhowmmConfig
import HhowmmFilename
import HhowmmGrep
import HhowmmReminder
import HhowmmResults
import HhowmmDefs

data GUIContext = GUIContext
    { guiFrame :: Frame ()
    , guiSearchEntry :: TextCtrl ()
    , guiLimitChoice :: Choice ()
    , config :: ConfigInfo
    }

limit_days :: [(String, Int)]
limit_days = [("All", 0), ("Today", 1), ("Half year", 182), ("Year", 365)]

defaultConfigFile :: String
defaultConfigFile = ".hhowmmrc"

reminderBGColor :: Int -> Color
reminderBGColor pr
    | pr == priority_invisible || pr == priority_bottom     = color_grey
    | pr == 0                                               = color_yellow
    | pr < 0                                                = color_pink
    | otherwise                                             = color_white

reminderFGColor :: Char -> Color
reminderFGColor rType = case rType of
    '@' -> color_black
    '!' -> color_darkred
    '-' -> color_blue
    '~' -> color_darkgreen
    _   -> color_black

findAllHashtags :: FilePath -> IO [String]
findAllHashtags dir = do
    files <- getRecursiveContents dir
    results <- mapM (\file -> findTags <$> readFile file) files
    return $ (nub . concat) results
  where
    findTags str = nub $ filter (\w -> isPrefixOf "#" w && length w > 1) $ words str

loadReminders :: FilePath -> IO [ReminderItem]
loadReminders dir = do
    files <- getRecursiveContents dir
    result <- searchReminders files
    return $ sort result

addReminderItem :: ListCtrl a -> ReminderItem -> IO ()
addReminderItem list (ReminderItem (GrepItem _ _ str) (ReminderInfo _ _ tp pr)) = do
    li <- listItemCreate
    listItemSetText li str
    listItemSetBackgroundColour li (reminderBGColor pr)
    listItemSetTextColour li (reminderFGColor tp)

    count <- listCtrlGetItemCount list
    listItemSetId li count
    listCtrlInsertItem list li
    return ()

openSearch :: GUIContext -> IO ()
openSearch (GUIContext _ ent cho conf) = do
    f <- getRecursiveContents $ confBaseDir conf
    let files = filterFilenameByExts f $ confExtensions conf
    query   <- get ent text
    c_idx   <- get cho selection
    let lim =  snd $ limit_days !! c_idx
    if lim == 0
        then resultsWindow conf (words query) files
        else do
            limitedFiles <- filterFilenameByDays files lim
            resultsWindow conf (words query) limitedFiles

gui :: ConfigInfo -> IO ()
gui conf = do
    f <- frame [text := "Hhowmm"]
    searchEntry     <- entry f []
    tagButton       <- button f [text := "#hashtags"]
-- ~     readButton      <- button f [text := "Read"]
    writeButton     <- button f [ text          := "Add new"
                                , on escKey     := windowSetFocus searchEntry
                                , on command    := do
                                    (d, file) <- newFilename
                                    let dir = (confBaseDir conf) </> d
                                    createDirectoryIfMissing True dir
                                    createProcess (shell $ (confEditorCmd conf) ++ " " ++ dir </> file)
                                    return ()
                                ]
    limitChoice   <- choice f [items := map fst limit_days, on escKey := windowSetFocus searchEntry]
    choiceSetSelection limitChoice 0
    let context     =  GUIContext f searchEntry limitChoice conf
    set searchEntry [ on escKey := do
                        query <- get searchEntry text
                        if null query
                            then close f
                            else set searchEntry [text := ""]
                    , on enterKey    := openSearch context]
-- ~     set readButton  [on command     := openSearch context]

    tags            <- findAllHashtags (confBaseDir conf)
    set tagButton   [ on escKey := windowSetFocus searchEntry
                    , on command := do
                        dObj        <- dialog f [text := "Choose a hashtag", clientSize := sz 100 250]
                        lBox        <- singleListBox dObj [items := tags]
                        cnxButton   <- button dObj [text := "Cancel"]
                        okButton    <- button dObj [text := "Ok"]
                        set dObj [layout :=
                                    margin 10 $
                                    column 2 [ hfill $ widget lBox
                                             , row 4 [hfill $ widget cnxButton, hfill $ widget okButton]
                                             ]
                                 ]
                        result <- showModal dObj (\stop -> do
                                                    set cnxButton [on command := stop Nothing, on escKey := stop Nothing]
                                                    set okButton  [on command := do
                                                                    num <- listBoxGetSelection lBox
                                                                    stop (Just num)
                                                                    ]
                                                    set lBox [ on escKey            := stop Nothing
                                                             , on (key KeySpace)    := do
                                                                num <- listBoxGetSelection lBox
                                                                stop (Just num)
                                                             ]
                                                    )
                        unless (isNothing result) $ do
                            tag <- listBoxGetString lBox (fromMaybe 0 result)
                            set searchEntry [text := tag]
                            openSearch context
                    ]

    remindersView   <- listCtrl f [ columns     := [("", AlignLeft, 400)]
                                  , font        := fontDefault{_fontSize = confFontSize conf}
                                  , style       :~ (.|. (wxLC_NO_HEADER .|. wxLC_SINGLE_SEL))
                                  , on escKey   := windowSetFocus searchEntry
                                  ]
    reminders <- loadReminders $ confBaseDir conf
    mapM_ (addReminderItem remindersView) reminders

    set f [layout :=
            margin 10 $
            column 2
                [ row 4 [hfill $ widget searchEntry, hfill $ widget tagButton]
                , hfill $ widget writeButton
-- ~                 , row 4 [hfill $ widget readButton, hfill $ widget writeButton]
                , hfill $ widget limitChoice
                , fill  $ widget remindersView
                ]
          , clientSize := sz 400 800
          ]
    windowSetFocus searchEntry
    return ()

main :: IO ()
main = do
    args <- getArgs
    homeDir <- getHomeDirectory
    let confFile = case args of
                [] -> homeDir </> defaultConfigFile
                _  -> head args
-- ~     acc <- fileAccess confFile True False False
    acc <- fileExist confFile
    conf <-
        if acc
            then readConfig confFile
            else return defaultConfig

    start $ gui conf
